# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import ClassVar, Iterator, Sequence

from babelwidget.main import (
    backend_t,
    grid_lyt_h,
    group_wgt_h,
    label_wgt_h,
    stack_wgt_h,
)

from conf_ini_g.interface.window.component.parameter.main import parameter_t
from conf_ini_g.interface.window.generic import FormattedName
from conf_ini_g.phase.instance.parameter.main import instance_t
from conf_ini_g.phase.specification.parameter.main import (
    parameter_t as parameter_spec_t,
)
from conf_ini_g.phase.specification.section.main import controller_t
from conf_ini_g.phase.specification.section.main import section_t as section_spec_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class _base_t:  # Cannot be abstracted
    HEADER_NAMES: ClassVar[tuple[str]] = (
        "Parameter",
        "Type(s)",
        "Value",
        "Unit",
    )
    HEADER_STYLE: ClassVar[str] = "background-color: darkgray; padding-left: 5px;"

    library_wgt: group_wgt_h
    formatted_name: str

    @classmethod
    def NewWithName(
        cls, name: str, backend: backend_t, /, *, controller: controller_t = None
    ) -> _base_t:
        """"""
        if controller is None:
            controller = ""
        else:
            controller = (
                f" ⮜ {FormattedName(controller.section, ' ')}."
                f"{FormattedName(controller.parameter, ' ')}"
            )
        formatted_name = FormattedName(name, " ") + controller

        output = cls(library_wgt=backend.group_wgt_t(), formatted_name=formatted_name)
        output.library_wgt.setTitle(formatted_name)

        return output

    @classmethod
    def Headers(cls, backend: backend_t, /) -> Sequence[label_wgt_h]:
        """"""
        output = []

        for text in cls.HEADER_NAMES:
            header = backend.label_wgt_t(f'<font color="blue">{text}</font>')
            header.setStyleSheet(cls.HEADER_STYLE)
            output.append(header)

        return output

    @property
    def all_parameters(self) -> Sequence[dict[str, parameter_t]]:
        """"""
        raise NotImplementedError

    @property
    def active_parameters(self) -> dict[str, parameter_t]:
        """"""
        raise NotImplementedError

    def __contains__(self, key: str, /) -> bool:
        """"""
        return any(key in _set.keys() for _set in self.all_parameters)

    def __getitem__(self, key: str, /) -> parameter_t:
        """"""
        for parameter_set in self.all_parameters:
            if key in parameter_set:
                return parameter_set[key]

        raise KeyError(f"{key}: Not a parameter of section {self.formatted_name}.")

    def __iter__(self) -> Iterator[parameter_t]:
        """"""
        for parameter_set in self.all_parameters:
            for parameter in parameter_set:
                yield parameter


@dtcl.dataclass(slots=True, repr=False, eq=False)
class section_t(_base_t):
    parameters: dict[str, parameter_t] = dtcl.field(init=False, default=None)

    @classmethod
    def NewForSection(
        cls,
        section_spec: section_spec_t,
        instances: dict[str, instance_t],
        backend: backend_t,
        /,
    ) -> section_t | None:
        """"""
        output = cls.NewWithName(section_spec.name, backend)

        parameters, parameter_names, layout = _SectionParameters(
            section_spec, instances, backend
        )
        if parameters.__len__() == 0:
            return None

        output.parameters = dict(zip(parameter_names, parameters))

        for h_idx, header in enumerate(cls.Headers(backend)):
            layout.addWidget(header, 0, h_idx)
        output.library_wgt.setLayout(layout)

        return output

    @property
    def all_parameters(self) -> Sequence[dict[str, parameter_t]]:
        """"""
        return (self.parameters,)

    @property
    def active_parameters(self) -> dict[str, parameter_t]:
        """"""
        return self.parameters


@dtcl.dataclass(slots=True, repr=False, eq=False)
class controlled_section_t(_base_t):
    parameter_sets: list[dict[str, parameter_t]] = dtcl.field(init=False, default=None)
    page_stack: stack_wgt_h = dtcl.field(init=False, default=None)

    @classmethod
    def NewForSection(
        cls,
        section_spec: section_spec_t,
        controller: controller_t,
        controller_value: str,
        instances: dict[str, instance_t],
        backend: backend_t,
        /,
    ) -> controlled_section_t | None:
        """"""
        output = cls.NewWithName(section_spec.name, backend, controller=controller)

        parameter_sets = []
        page_stack = backend.stack_wgt_t()
        for parameter_specs in (section_spec, *section_spec.alternatives.values()):
            parameters, parameter_names, layout = _SectionParameters(
                parameter_specs, instances, backend
            )
            if parameters.__len__() == 0:
                continue

            parameter_sets.append(dict(zip(parameter_names, parameters)))

            for h_idx, header in enumerate(cls.Headers(backend)):
                layout.addWidget(header, 0, h_idx)
            page = backend.library_wgt_t()
            page.setLayout(layout)
            page_stack.addWidget(page)

        if parameter_sets.__len__() == 0:
            return None

        output.parameter_sets = parameter_sets
        output.page_stack = page_stack

        controlling_values = section_spec.controlling_values
        page_stack.setCurrentIndex(controlling_values.index(controller_value))

        # Curiously, the stacked widget cannot be simply declared as child of instance;
        # This must be specified through a layout.
        layout = backend.hbox_lyt_t()
        layout.addWidget(page_stack)
        layout.setContentsMargins(0, 0, 0, 0)
        output.library_wgt.setLayout(layout)

        return output

    @property
    def all_parameters(self) -> Sequence[dict[str, parameter_t]]:
        """"""
        return self.parameter_sets

    @property
    def active_parameters(self) -> dict[str, parameter_t]:
        """"""
        return self.parameter_sets[self.page_stack.currentIndex()]


def _SectionParameters(
    specifications: section_spec_t | Iterator[parameter_spec_t],
    instances: dict[str, instance_t],
    backend: backend_t,
    /,
) -> tuple[Sequence[parameter_t], Sequence[str], grid_lyt_h]:
    """"""
    parameters = []
    parameter_names = []

    layout = backend.grid_lyt_t()
    layout.setAlignment(backend.ALIGNED_TOP)
    layout.setColumnStretch(0, 4)
    layout.setColumnStretch(1, 1)
    layout.setColumnStretch(2, 8)
    layout.setColumnStretch(3, 1)
    layout.setContentsMargins(0, 0, 0, 0)

    for row, parameter_spec in enumerate(specifications, start=1):
        parameter = parameter_t.NewForParameter(
            parameter_spec, instances[parameter_spec.name], backend
        )
        parameters.append(parameter)
        parameter_names.append(parameter_spec.name)

        layout.addWidget(parameter.name, row, 0, alignment=backend.ALIGNED_RIGHT)
        layout.addWidget(parameter.type, row, 1)
        layout.addWidget(parameter.value.library_wgt, row, 2, 1, 2 - 1)
        if parameter.unit is not None:
            layout.addWidget(parameter.unit, row, 3)

    return parameters, parameter_names, layout



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import textwrap as text
from typing import Any, Sequence

from rich.text import Text as text_t
from str_to_obj.interface.console import TypeAsRichStr

from conf_ini_g.extension.path import any_path_h, path_t
from conf_ini_g.extension.string import AlignedOnSeparator
from conf_ini_g.interface.constant import (
    INI_COMMENT_MARKER,
    INI_UNIT_SECTION,
    STD_UNIT_CONVERSIONS,
)
from conf_ini_g.phase.instance.parameter.main import instance_t
from conf_ini_g.phase.instance.parameter.unit import unit_instance_t
from conf_ini_g.phase.raw.config import ini_config_h, typed_config_h
from conf_ini_g.phase.specification.config.main import config_t as specification_t
from conf_ini_g.phase.specification.parameter.unit import unit_t
from conf_ini_g.phase.specification.parameter.value import MISSING_REQUIRED_VALUE
from conf_ini_g.phase.specification.section.main import controller_t
from conf_ini_g.phase.specification.section.unit import IsUnitSection


@dtcl.dataclass(slots=True, repr=False, eq=False)
class config_t(dict[str, dict[str, instance_t | unit_instance_t]]):
    specification: specification_t
    ini_path: path_t | None = None

    @classmethod
    def NewFromRawConfig(
        cls,
        ini_config: ini_config_h,
        specification: specification_t,
        /,
        *,
        path: any_path_h = None,
    ) -> tuple[config_t, list[str], Sequence[tuple[str, str]]]:
        """"""
        config = cls(specification=specification)

        if path is not None:
            config.ini_path = path_t(path)

        issues, for_deferred_check = config._InstantiateFromINIConfig(ini_config)
        config._ManageDefaults()

        return config, issues, for_deferred_check

    def _InstantiateFromINIConfig(
        self, ini_config: ini_config_h, /
    ) -> tuple[list[str], Sequence[tuple[str, str]]]:
        """"""
        issues = []
        for_deferred_check = []

        for section_name, parameters in ini_config.items():
            if section_name not in self:
                self[section_name] = {}

            is_specified = section_name in self.specification
            if (not is_specified) and IsUnitSection(section_name):
                self.specification.AddUnitSection()
                is_specified = True

            if is_specified:
                section_spec = self.specification[section_name]
                for name, value in parameters.items():
                    if name in section_spec:
                        parameter_spec = section_spec[name]
                    elif section_spec.is_growable:
                        if (controller := section_spec.controller) is None:
                            controlling_value = None
                        else:
                            controlling_value = self.GetValueOfController(controller)
                        parameter_spec = self.specification.AddINIParameter(
                            section_spec,
                            name,
                            value,
                            controlling_value=controlling_value,
                        )
                        for_deferred_check.append((section_name, name))
                    else:
                        parameter_spec = None
                        issues.append(
                            f"{section_name}.{name}: Attempt to add an unspecified "
                            f"parameter to a section accepting none."
                        )
                    if parameter_spec is not None:
                        if isinstance(parameter_spec, unit_t):
                            instance = unit_instance_t()
                        else:
                            instance = instance_t()
                        instance.SetINIorInterfaceOrDefaultValue(
                            value, INI_COMMENT_MARKER
                        )
                        self[section_name][name] = instance
            elif IsUnitSection(section_name, possibly_fuzzy=True):
                issues.append(
                    f"{section_name}: Unit section must respect the following case "
                    f'"{INI_UNIT_SECTION}".'
                )
            else:
                issues.append(
                    f"{section_name}: Invalid section; "
                    f"Expected={self.specification.section_names}."
                )

        return issues, for_deferred_check

    def _ManageDefaults(self) -> None:
        """"""
        for section_spec in self.specification:
            # Leave here, otherwise parameter_spec.name not in self[section_spec.name]
            # below might fail.
            if section_spec.name not in self:
                self[section_spec.name] = {}

            for parameter_spec in section_spec.all_parameters:
                if parameter_spec.name not in self[section_spec.name]:
                    if parameter_spec.optional:
                        instance = instance_t(value=parameter_spec.default)
                        self[section_spec.name][parameter_spec.name] = instance
                    else:
                        self[section_spec.name][
                            parameter_spec.name
                        ] = MISSING_REQUIRED_VALUE

    def GetValueOfController(self, controller: controller_t, /) -> Any:
        """"""
        return self[controller.section][controller.parameter].value

    def AsTypedConfig(self) -> tuple[typed_config_h, list[str]]:
        """
        Units are interpreted
        """
        typed_config = {}
        issues = []

        unit_conversions = dict(STD_UNIT_CONVERSIONS)
        if INI_UNIT_SECTION in self.specification:
            for unit_spec in self.specification[INI_UNIT_SECTION]:
                unit_name = unit_spec.name
                expected_type = unit_spec.type
                instance = self[INI_UNIT_SECTION][unit_name]
                if instance is MISSING_REQUIRED_VALUE:
                    issues.append(
                        f"{INI_UNIT_SECTION}.{unit_name}: Missing required unit."
                    )
                else:
                    value, current_issues = instance.TypedValue(expected_type)
                    if current_issues.__len__() > 0:
                        issues.extend(
                            f"/{INI_UNIT_SECTION}.{unit_name}/ {_iss}"
                            for _iss in current_issues
                        )
                    else:
                        unit_conversions[unit_name] = value

        for section_spec in self.specification:
            raw_section = {}

            # Note: Method interface.window.config.config_t.SectionActiveParameterSpec
            # is defined using a similar piece of code.
            if (controller := section_spec.controller) is None:
                parameters = section_spec
            else:
                controller_value = self.GetValueOfController(controller)
                parameters = section_spec.ActiveParameters(controller_value)
            for parameter_spec in parameters:
                instance = self[section_spec.name][parameter_spec.name]
                if instance is MISSING_REQUIRED_VALUE:
                    value = None
                    current_issues = ["Missing required parameter."]
                else:
                    value, current_issues = instance.TypedValue(
                        parameter_spec.type, units=unit_conversions
                    )

                if current_issues.__len__() > 0:
                    issues.extend(
                        f"/{section_spec.name}.{parameter_spec.name}/ {_iss}"
                        for _iss in current_issues
                    )
                else:
                    raw_section[parameter_spec.name] = value

            typed_config[section_spec.name] = raw_section

        return typed_config, issues

    def AsINIConfig(self) -> ini_config_h:
        """"""
        output = {}

        for section_name, section in self.items():
            section_spec = self.specification[section_name]
            if (controller := section_spec.controller) is None:
                raw_section = {_p_nme: _prm.Text() for _p_nme, _prm in section.items()}
            else:
                controller_value = self.GetValueOfController(controller)
                raw_section = {
                    _p_spec.name: section[_p_spec.name].Text()
                    for _p_spec in section_spec.ActiveParameters(controller_value)
                }

            output[section_name] = raw_section

        return output

    def Issues(self) -> list[str]:
        """"""
        output = self.specification.Issues()

        valid_units = list(STD_UNIT_CONVERSIONS.keys())
        if INI_UNIT_SECTION in self.specification:
            valid_units.extend(
                _unt.name for _unt in self.specification[INI_UNIT_SECTION]
            )

        for section in self.specification:
            section_name = section.name

            issues = []
            for parameter in section:
                if (instance := self[section.name].get(parameter.name)) is not None:
                    issues.extend(
                        f"/{section.name}.{parameter.name}/ {_iss}"
                        for _iss in instance.Issues()
                    )
                    # Unit validation cannot be done at the parameter level since the
                    # unit section is needed.
                    if ((unit := getattr(instance, "unit", None)) is not None) and (
                        unit not in valid_units
                    ):
                        issues.append(
                            f"{unit}: Invalid unit of parameter "
                            f"{section_name}.{parameter.name}"
                        )

            output.extend(issues)

        return output

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [
            "[grey50]--- Specification[/]",
            TypeAsRichStr(self),
            f"    [blue]ini_path[/]={self.ini_path}"
            f"[yellow]:{type(self.ini_path).__name__}[/]",
            text.indent(self.specification.__rich__(), "    "),
            "[grey50]--- Instance[/]",
        ]

        for section_name, section in self.items():
            output.append(f"{section_name}:")
            for parameter_name, parameter in section.items():
                output.append(f"    {parameter_name}@:@{parameter.__rich__()}")

        output = AlignedOnSeparator(output, "@:@", ": ")

        return "\n".join(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from abc import ABC as abstract_class_t
from abc import abstractmethod
from typing import Any

from babelwidget.main import backend_t
from str_to_obj.type.annotation import annotation_t

from conf_ini_g.phase.specification.parameter.main import parameter_t as specification_t
from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class parameter_a(abstract_class_t):
    @classmethod
    @abstractmethod
    def NewWithDetails(
        cls,
        value: Any | None,
        value_type: type_t | annotation_t | None,
        backend: backend_t,
        specification: specification_t | None,
    ) -> parameter_a:
        ...

    @abstractmethod
    def Assign(self, value: Any, value_type: type_t | annotation_t | None, /) -> None:
        ...

    @abstractmethod
    def Text(self) -> str:
        ...



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Any, Sequence

from babelwidget.main import backend_t, choices_list_wgt_h, library_wgt_h, stack_wgt_h
from str_to_obj.type.hint_tree import hint_tree_t

from conf_ini_g.catalog.interface.window.parameter.directory import WidgetTypeForType


@dtcl.dataclass(slots=True, repr=False, eq=False)
class multitype_wgt_t:
    hints: Sequence[hint_tree_t]
    library_wgt: stack_wgt_h
    controller_wgt: choices_list_wgt_h
    values: tuple[library_wgt_h] | None = dtcl.field(init=False, default=None)

    @classmethod
    def NewForHints(
        cls,
        value: Any,
        hints: Sequence[hint_tree_t],
        controller: choices_list_wgt_h,
        backend: backend_t,
        /,
    ) -> multitype_wgt_t:
        """"""
        value_stack = backend.stack_wgt_t()
        controller.SetFunction(value_stack.setCurrentIndex)

        output = cls(hints=hints, library_wgt=value_stack, controller_wgt=controller)

        values = []
        for t_idx, hint in enumerate(hints):
            if isinstance(value, hint.type):
                initial_value = value
            else:
                initial_value = None
            widget_type = WidgetTypeForType(hint)
            value_wgt = widget_type.NewWithDetails(
                initial_value,
                hint,
                backend,
                None,
            )
            values.append(value_wgt)
            value_stack.addWidget(value_wgt.library_wgt)
        output.values = tuple(values)

        value_stack.setSizePolicy(backend.SIZE_EXPANDING, backend.SIZE_FIXED)

        output.Assign(value, None)

        return output

    def Assign(self, value: Any, _: Any | None, /) -> None:
        """
        Second parameter: For coherence with single-type parameters.
        """
        initial_index = 0
        for t_idx, hint in enumerate(self.hints):
            if isinstance(value, hint.type):
                initial_index = t_idx
        self.controller_wgt.setCurrentIndex(initial_index)

    def Text(self) -> str:
        """"""
        return self.values[self.controller_wgt.currentIndex()].Text()



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
from __future__ import annotations

from typing import Callable, Sequence

from babelwidget.main import backend_t, button_wgt_h, grid_lyt_h

from conf_ini_g.phase.raw.config import typed_config_h


def ActionButtonsLayout(
    action: tuple[str, Callable[[typed_config_h], None]] | None,
    has_ini_document: bool,
    backend: backend_t,
    ShowInINIFormat: Callable,
    SaveConfig: Callable,
    LaunchAction: Callable,
    Close: Callable,
    /,
) -> tuple[grid_lyt_h, button_wgt_h, Sequence[button_wgt_h]]:
    """"""
    layout = backend.grid_lyt_t()

    buttons = []
    geometries = []

    button = backend.button_wgt_t("Show in INI format")
    button.SetFunction(ShowInINIFormat)
    buttons.append(button)
    geometries.append((0, 0, 1, 2))

    button = backend.button_wgt_t("Save Config As")
    button.SetFunction(lambda: SaveConfig(True))
    buttons.append(button)
    if has_ini_document:
        geometries.append((1, 0, 1, 1))

        button = backend.button_wgt_t("Save/Overwrite Config")
        button.SetFunction(lambda: SaveConfig(False))
        buttons.append(button)
        geometries.append((1, 1, 1, 1))
    else:
        geometries.append((1, 0, 1, 2))

    if action is None:
        label = "CLOSE"
        Function = Close
    else:
        label = action[0]
        Function = LaunchAction

    button = backend.button_wgt_t(label)
    button.SetFunction(Function)
    buttons.append(button)
    geometries.append((2, 0, 1, 2))

    action_button = button

    for button, geometry in zip(buttons, geometries):
        layout.addWidget(button, *geometry)
    layout.setContentsMargins(0, 0, 0, 0)

    return layout, action_button, buttons



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import inspect as nspt
import sys as sstm
from typing import Callable

from rich.text import Text as text_t
from str_to_obj.interface.console import NameValueTypeAsRichStr, TypeAsRichStr

from conf_ini_g.extension.path import any_path_h, path_t
from conf_ini_g.extension.string import AlignedOnSeparator
from conf_ini_g.phase.raw.config import ini_config_h
from conf_ini_g.phase.specification.parameter.type import type_t

_SECTION_PARAMETER_SEPARATOR = "__"


class config_t:
    path: path_t
    _types: dict[str, type_t]

    @classmethod
    def NewFromDictionary(
        cls, ini_config: ini_config_h, /, *, path: any_path_h = None
    ) -> config_t:
        """
        The dictionary values are already properly typed
        """
        output = cls()

        if (path is not None) and isinstance(path, str):
            path = path_t(path)

        output.path = path
        output._types = {
            _nme: type_t.NewFromTypeHint(_hnt)
            for _nme, _hnt in nspt.get_annotations(cls).items()
        }

        issues = []
        for s_name, section in ini_config.items():
            for p_name, value in section.items():
                issues.extend(output.Set(_FullName(s_name, p_name), value))

        if issues.__len__() > 0:
            issues = "\n".join(issues)
            print(f"{path}: Invalid configuration.\n{issues}")
            sstm.exit(1)

        return output

    def Set(self, name: str, value: str, /) -> list[str]:
        """"""
        expected_type = self._types[name]
        typed_value, issues = expected_type.InterpretedValueOf(value)

        if issues.__len__() > 0:
            return [
                f"{value}: "
                f'Invalid value of parameter "{name}": {", ".join(issues)}.'
            ]

        setattr(self, name, typed_value)
        return []

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [TypeAsRichStr(self)]

        AllButCallable = lambda _elm: not isinstance(_elm, Callable)
        for name, value in nspt.getmembers(self, AllButCallable):
            if name[0] != "_":
                name = name.replace(_SECTION_PARAMETER_SEPARATOR, ".", 1)
                output.append(
                    "    " + NameValueTypeAsRichStr(name, value, separator="@=@")
                )

        output = AlignedOnSeparator(output, "@=@", " = ")

        return "\n".join(output)


def _FullName(section: str, parameter: str, /) -> str:
    """"""
    return f"{section.strip().lower().replace(' ', '_')}{_SECTION_PARAMETER_SEPARATOR}{parameter.strip().lower()}"



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import ClassVar, Sequence

from babelwidget.main import backend_t, choices_list_wgt_h, library_wgt_h, text_wgt_h

from conf_ini_g.catalog.specification.parameter.sequence import sequence_t
from conf_ini_g.catalog.specification.parameter.sequence import (
    sequence_t as sequence_annotation_t,
)
from conf_ini_g.interface.window.component.parameter.protocol import parameter_a
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.parameter.type import type_t

VALUE_FOR_NONE = ()


@dtcl.dataclass(repr=False, eq=False)
class sequence_wgt_t(parameter_a):
    """
    Cannot use slots (weak reference issue).
    """

    ENTRY_ANY: ClassVar[str] = "any"
    ENTRIES: ClassVar[tuple[str, ...]] = ("2", "3", "4", "5", "6", ENTRY_ANY)

    library_wgt: library_wgt_h
    backend: backend_t
    delimiters: str = None
    length_selector: choices_list_wgt_h = dtcl.field(init=False, default=None)
    components: tuple[text_wgt_h, ...] = dtcl.field(init=False, default=None)

    @classmethod
    def NewWithDetails(
        cls,
        value: tuple | None,
        value_type: type_t | sequence_annotation_t | None,
        backend: backend_t,
        _: parameter_t | None,
        /,
    ) -> sequence_wgt_t:
        """
        If value_type does not contain the necessary details, a simple free-text input widget is used. If the value is not
        coherent with the details (which should not happen if value_type contains the necessary details and the value has
        been validated), a choice with the length of the value is added, with an exclamation point.
        """
        output = cls(library_wgt=backend.library_wgt_t(), backend=backend)

        if value is None:
            output.delimiters = "()"
        elif isinstance(value, tuple):
            output.delimiters = "()"
        elif isinstance(value, list):
            output.delimiters = "[]"
        else:
            raise TypeError(f"{type(value).__name__}: Unhandled sequence type.")

        output.Assign(value, value_type)

        layout = backend.hbox_lyt_t()
        layout.setContentsMargins(0, 0, 0, 0)
        if output.length_selector is not None:
            layout.addWidget(output.length_selector)
        for component in output.components:
            layout.addWidget(component)
        output.library_wgt.setLayout(layout)

        return output

    def SetLength(self, new_index: int) -> None:
        """"""
        new_length = self.length_selector.ItemAt(new_index)
        if new_length == self.__class__.ENTRY_ANY:
            new_length = 1
        elif new_length.endswith("!"):
            new_length = int(new_length[:-1])
        else:
            new_length = int(new_length)
        _AdjustComponents(self.components, new_length)

    def Assign(
        self, value: tuple | None, value_type: type_t | sequence_annotation_t | None, /
    ) -> None:
        """"""
        cls = self.__class__

        if value is None:
            value = VALUE_FOR_NONE

        length = value.__len__()

        default_parameters = (cls.ENTRIES, int(cls.ENTRIES[-2]))
        if value_type is None:
            entries, max_entry = default_parameters
        else:
            if isinstance(value_type, type_t):
                annotation = value_type.FirstAnnotationWithAttribute("lengths")
            else:
                annotation = value_type
            if annotation is None:
                entries, max_entry = default_parameters
            elif (lengths := annotation.lengths) == sequence_t.ANY_LENGTH:
                entries = (cls.ENTRY_ANY,)
                max_entry = None
            else:
                entries = tuple(str(_lgh) for _lgh in lengths)
                max_entry = lengths[-1]
                if (length > 0) and (str(length) not in entries):
                    # This should never happen since the value must have been validated
                    entries = entries + (str(length) + "!",)
                    max_entry = max(max_entry, length)

        components = []
        if max_entry is None:
            value_as_str = str(value)[1:-1] if length > 0 else ""
            if self.components is None:
                widget = self.backend.text_wgt_t(value_as_str, parent=None)
                components.append(widget)
            else:
                self.components[0].setText(value_as_str)
        else:
            for e_idx in range(max_entry):
                value_as_str = str(value[e_idx]) if e_idx < length else ""
                if self.components is None:
                    widget = self.backend.text_wgt_t(value_as_str, parent=None)
                    components.append(widget)
                else:
                    widget = self.components[e_idx]
                if e_idx >= length:
                    widget.setVisible(False)
                    widget.setEnabled(False)
        if self.components is None:
            self.components = tuple(components)

        if entries.__len__() > 1:
            if self.length_selector is None:
                length_selector = self.backend.choice_menu_wgt_t()
                for entry in entries:
                    length_selector.addItem(entry)
            else:
                length_selector = self.length_selector

            if (length_as_str := str(length)) in entries:
                length_selector.setCurrentText(length_as_str)
            if length_as_str + "!" in entries:
                length_selector.setCurrentText(length_as_str + "!")
            else:  # cls.ENTRY_ANY is necessarily in entries
                length_selector.setCurrentText(cls.ENTRY_ANY)

            if self.length_selector is None:
                self.length_selector = length_selector
                self.length_selector.SetFunction(self.SetLength)

    def Text(self) -> str:
        """"""
        contents = []
        for component in self.components:
            # Do not use "visible" here since setting visible does not really set the property until it is actually
            # shown. The documentation explains about ancestors being visible or not, but it was not clear that the
            # property is apparently not effective immediately.
            if not component.isEnabled():
                break
            component_as_str = component.Text()
            if component_as_str == "":
                break
            contents.append(component_as_str)

        if contents.__len__() == 0:
            return self.delimiters
        elif contents.__len__() == 1:
            if self.delimiters[0] == "(":
                return f"({contents[0]},)"
            else:
                return f"[{contents[0]}]"
        else:
            return self.delimiters[0] + ", ".join(contents) + self.delimiters[1]


def _AdjustComponents(
    components: Sequence[text_wgt_h, ...],
    length: int,
) -> None:
    """"""
    for c_idx, component in enumerate(components):
        if c_idx < length:
            component.setVisible(True)
            component.setEnabled(True)
        else:
            component.setVisible(False)
            component.setEnabled(False)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from typing import Any, ClassVar, Sequence

from str_to_obj import annotation_t
from str_to_obj.type.hint import any_hint_h

from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(repr=False, eq=False)
class sequence_t(annotation_t):
    ANY_ITEMS_TYPES: ClassVar[Any | tuple[Any, ...]] = Any
    ANY_LENGTH: ClassVar[tuple[int, ...]] = (0,)

    ACCEPTED_TYPES = (list, set, tuple)
    # Any=Value of any type but None
    items_types: any_hint_h | tuple[any_hint_h, ...] | type_t | tuple[
        type_t, ...
    ] = ANY_ITEMS_TYPES
    lengths: int | tuple[int, ...] = ANY_LENGTH

    def __post_init__(self) -> None:
        """"""
        original_item_types = self.items_types
        original_lengths = self.lengths

        if isinstance(self.items_types, Sequence):
            items_types = []
            for type_ in self.items_types:
                if isinstance(type_, type_t):
                    items_types.append(type_)
                else:
                    items_types.append(type_t.NewFromTypeHint(type_))
            self.items_types = items_types
        elif not isinstance(self.items_types, type_t):
            self.items_types = type_t.NewFromTypeHint(self.items_types)

        if isinstance(self.lengths, int):
            self.lengths = (self.lengths,)
        else:
            self.lengths = tuple(sorted(self.lengths))

        if isinstance(self.items_types, Sequence):
            if max(self.lengths) > self.items_types.__len__():
                raise ValueError(
                    f"{original_item_types}/{original_lengths}: Allowed length(s) must "
                    f"not exceed the length of the item types sequence."
                )

    def Issues(self) -> list[str]:
        """"""
        output = []

        if self.lengths != self.__class__.ANY_LENGTH:
            for length in self.lengths:
                if (not isinstance(length, int)) or (length < 0):
                    output.append(
                        f"{length}: Invalid sequence length in {self}; "
                        f"Expected=strictly positive integer"
                    )

        return output

    def ValueIsCompliant(self, value: tuple, /) -> list[str]:
        """"""
        if (self.lengths != self.__class__.ANY_LENGTH) and (
            value.__len__() not in self.lengths
        ):
            return [
                f"{value}: Sequence of invalid length; "
                f"Expected={' or '.join(map(str, self.lengths))}."
            ]

        if isinstance(self.items_types, Sequence):
            if value.__len__() > self.items_types.__len__():
                return [
                    f"{value}: Sequence too long. Expected=At most {self.items_types.__len__()} element(s)."
                ]

            output = []
            for type_, element in zip(self.items_types, value):
                output.extend(type_.ValueIsCompliant(element))
            return output

        type_: type_t = self.items_types
        output = []
        for element in value:
            output.extend(type_.ValueIsCompliant(element))
        return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re as rgex
from argparse import ArgumentParser as argument_parser_t
from pathlib import Path as path_t
from typing import Any, Iterator, Sequence

from conf_ini_g.extension.string import Flattened
from conf_ini_g.phase.raw.config import typed_config_h
from conf_ini_g.phase.specification.config.main import config_t
from conf_ini_g.phase.specification.parameter.value import MISSING_REQUIRED_VALUE

parsed_arguments_h = dict[str, str]


# Specified INI document file is stored in _INI_DOCUMENT_VARIABLE
_INI_DOCUMENT_VARIABLE = "ini_path"

_ADVANCED_MODE_OPTION = "advanced-mode"
_ADVANCED_MODE_VARIABLE = "advanced_mode"

# Usage: {section_name}{_SECTION_PARAMETER_SEPARATOR}{parameter_name}
_SECTION_PARAMETER_SEPARATOR = "-"


def CommandLineParser(
    description: str | None, specification: config_t, /
) -> argument_parser_t:
    """"""
    output = argument_parser_t(description=description, allow_abbrev=False)
    # This argument is optional because all parameters can be explicitly passed as
    # arguments (CLI mode), or nothing can be passed at all (GUI mode).
    output.add_argument(
        dest=_INI_DOCUMENT_VARIABLE,
        help="Path to INI configuration file",
        default=None,
        nargs="?",
        metavar="INI_config_file",
    )

    options = []
    for section in specification:
        for parameter in section.all_parameters:
            option = f"{section.name}{_SECTION_PARAMETER_SEPARATOR}{parameter.name}"
            if option == _ADVANCED_MODE_OPTION:
                # Raising an exception is adapted here since execution has been launched from command line
                raise ValueError(
                    f"{option}: Command-line option for parameter is identical to advanced mode option. "
                    f"Please change parameter specification (section name and/or parameter name)."
                )
            if option in options:
                # This can happen in controlled sections whose alternatives share equally named parameters.
                continue

            attribute = f"{section.name}{_SECTION_PARAMETER_SEPARATOR}{parameter.name}"

            # Default is a missing_required_value_t instance to avoid overwriting if in
            # INI but not passed.
            default = MISSING_REQUIRED_VALUE

            if parameter.optional:
                if isinstance(parameter.default, str):
                    delimiter = '"'
                else:
                    delimiter = ""
                type_and_value = (
                    f"Type: {parameter.type}. "
                    f"Default: [green]{delimiter}{parameter.default}{delimiter}[/]"
                )
            else:
                type_and_value = str(default)
            flattened = Flattened(type_and_value)
            definition = f"{parameter.definition}. {flattened}"

            # Type could be TypeOfAnnotatedHint(cmd_line_type). However, to allow passing any of the allowed types,
            # deferring type validation to functional config instantiation, this parameter is not passed.
            output.add_argument(
                f"--{option}",
                dest=attribute,
                help=definition,
                default=default,
                metavar=attribute,
            )
            options.append(option)

    output.add_argument(
        f"--{_ADVANCED_MODE_OPTION}",
        dest=_ADVANCED_MODE_VARIABLE,
        help="Toggle display of advanced sections and parameters",
        action="store_true",
    )

    return output


def ParsedArguments(
    parser: argument_parser_t, /, *, arguments: Sequence[str] = None
) -> tuple[path_t | None, bool, typed_config_h]:
    """"""
    parsed, unknowns = parser.parse_known_args(args=arguments)
    parsed = vars(parsed)

    advanced_mode = parsed[_ADVANCED_MODE_VARIABLE]
    del parsed[_ADVANCED_MODE_VARIABLE]

    ini_path = parsed[_INI_DOCUMENT_VARIABLE]
    del parsed[_INI_DOCUMENT_VARIABLE]
    if ini_path is not None:
        try:
            ini_path = path_t(ini_path).resolve(strict=True)
        except (FileNotFoundError, RuntimeError):
            raise FileNotFoundError(f"Nonexistent or invalid path: {ini_path}")

    pattern = rgex.compile(
        r"--(\w+)" + _SECTION_PARAMETER_SEPARATOR + r"(\w+)=(.+)", flags=rgex.ASCII
    )
    for unknown in unknowns:
        match = pattern.fullmatch(unknown)
        if match is None:
            # Raising an exception is adapted here since execution has been launched from command line
            raise ValueError(
                f"{unknown}: Invalid option syntax; Expected=--SECTION-PARAMETER=VALUE"
            )

        section, parameter, value = match.groups()
        parsed[f"{section}{_SECTION_PARAMETER_SEPARATOR}{parameter}"] = value

    raw_config = {}
    for sct_name, prm_name, value in _SectionParameterValueIterator(parsed):
        if sct_name in raw_config:
            raw_config[sct_name][prm_name] = value
        else:
            raw_config[sct_name] = {prm_name: value}

    return ini_path, advanced_mode, raw_config


def _SectionParameterValueIterator(
    arguments: parsed_arguments_h,
) -> Iterator[tuple[str, str, Any]]:
    """"""
    for prm_uid, value in arguments.items():
        # See CommandLineParser for why this can happen.
        if value is MISSING_REQUIRED_VALUE:
            continue

        section, parameter = prm_uid.split(_SECTION_PARAMETER_SEPARATOR)
        yield section, parameter, value



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from pathlib import Path as pl_path_t
from typing import Callable

from babelwidget.main import backend_t, library_wgt_h, text_wgt_h
from babelwidget.path_chooser import (
    NewSelectedInputDocument,
    NewSelectedOutputDocument,
    document_selection_fct_h,
)

from conf_ini_g.catalog.specification.parameter.path import path_t
from conf_ini_g.interface.constant import PATH_SELECTOR_WIDTH
from conf_ini_g.interface.window.component.parameter.protocol import parameter_a
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(repr=False, eq=False)
class path_wgt_t(parameter_a):
    """
    Cannot use slots (weak reference issue).
    """

    PostAssignmentFunction: Callable[[pl_path_t], None]
    library_wgt: library_wgt_h
    editable: bool
    backend: backend_t
    target_type: path_t.TARGET_TYPE = dtcl.field(init=False, default=None)
    path: text_wgt_h = dtcl.field(init=False, default=None)
    _NewSelectedDocument: document_selection_fct_h = dtcl.field(
        init=False, default=None
    )

    @classmethod
    def NewWithDetails(
        cls,
        value: pl_path_t | None,
        value_type: type_t | path_t | None,
        backend: backend_t,
        _: parameter_t | None,
        /,
        *,
        editable: bool = True,
        PostAssignmentFunction: Callable[[pl_path_t], None] = None,
    ) -> path_wgt_t:
        """
        If value_type does not contain the necessary details, the target type is set to any and considered as input, and
        the selection button label ends with an exclamation point.
        """
        output = cls(
            PostAssignmentFunction=PostAssignmentFunction,
            library_wgt=backend.library_wgt_t(),
            editable=editable,
            backend=backend,
        )

        value = _ValidStrValue(value, editable)

        default_parameters = (path_t.TARGET_TYPE.any, True, True)
        if value_type is None:
            target_type, is_input, misses_details = default_parameters
        else:
            if isinstance(value_type, type_t):
                annotation = value_type.FirstAnnotationWithAttribute(
                    ("target_type", "is_input")
                )
            else:
                annotation = value_type
            if annotation is None:
                target_type, is_input, misses_details = default_parameters
            else:
                target_type = annotation.target_type
                is_input = annotation.is_input
                misses_details = False

        output.target_type = target_type
        if is_input:
            output._NewSelectedDocument = NewSelectedInputDocument
        else:
            output._NewSelectedDocument = NewSelectedOutputDocument

        if target_type is path_t.TARGET_TYPE.document:
            selector_label = "🗋"
        elif target_type is path_t.TARGET_TYPE.folder:
            selector_label = "📂"
        else:
            selector_label = "📂🗋"
        if misses_details:
            selector_label += " !"
        if is_input:
            selector_color = "green"
        else:
            selector_color = "red"
        if editable:
            path = backend.text_wgt_t(value, parent=output.library_wgt)
        else:
            path = backend.label_wgt_t(value, parent=output.library_wgt)
        path_selector = backend.button_wgt_t(selector_label, parent=output.library_wgt)
        path_selector.SetFunction(output.SelectDocument)

        output.path = path

        path_selector.setStyleSheet(f"color: {selector_color};")
        path_selector.setFixedWidth(PATH_SELECTOR_WIDTH)

        layout = backend.hbox_lyt_t()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(path)
        layout.addWidget(path_selector)
        output.library_wgt.setLayout(layout)

        return output

    def Assign(self, value: pl_path_t | None, _: type_t | path_t | None, /) -> None:
        """"""
        value = _ValidStrValue(value, self.editable)
        self.path.setText(value)

    def Text(self) -> str:
        """"""
        return self.path.Text()

    def SelectDocument(self) -> None:
        """"""
        current_path = self.Text()
        current_doc = pl_path_t(current_path).resolve()

        if self.target_type is path_t.TARGET_TYPE.document:
            title = "Select File"
        elif self.target_type is path_t.TARGET_TYPE.folder:
            title = "Select Folder"
        else:
            title = "Select File or Folder"

        selection = self._NewSelectedDocument(
            title,
            title,
            self.backend,
            mode=self.target_type.name,
            start_folder=current_doc.parent,
            initial_selection=current_doc,
        )
        if selection is None:
            return

        self.Assign(selection, None)
        # Put post-assignment call here instead of in the Assign method in case Assign
        # is also called at initialization time, in NewWithDetails, one day. Indeed, the
        # post-assignment task should have been done already, or will be done, when
        # instantiating the interface.
        if self.PostAssignmentFunction is not None:
            self.PostAssignmentFunction(selection)


def _ValidStrValue(value: pl_path_t | None, editable: bool, /) -> str:
    """"""
    if value is None:
        output = ""
    else:
        if not editable:
            value = value.relative_to(pl_path_t.home())
        output = str(value)

    return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import ast as bstr
import inspect as nspt
import sys as sstm
from pathlib import Path as path_t


def SpecificationPath(element: object, /, *, relative_to_home: bool = True) -> str:
    """"""
    try:
        found = [nspt.getfile(element)]
    except TypeError:
        found = []
    if found.__len__() == 0:
        for module in sstm.modules.copy().values():
            for attribute in dir(module):
                # It can happen that "attribute" ends up not being an attribute of
                # "module", for example with "single" in scipy.linalg.matfuncs. Hence,
                # the default value in getattr. The getattr call can even fail, for
                # example with "Viewer" in napari. Hence the try-expect.
                try:
                    value = getattr(module, attribute, None)
                except:
                    value = None
                if (value is not None) and (value is element):
                    code = nspt.getsource(module)
                    tree = bstr.parse(code)
                    is_imported = False
                    for node in bstr.walk(tree):
                        if isinstance(node, bstr.ImportFrom):
                            for alias in node.names:
                                if (
                                    (alias.asname is None) and (alias.name == attribute)
                                ) or (
                                    (alias.asname is not None)
                                    and (alias.asname == attribute)
                                ):
                                    is_imported = True
                                    break
                    if not is_imported:
                        path = module.__file__
                        if relative_to_home:
                            path = str(path_t(path).relative_to(path_t.home()))
                        found.append(path)

    if (n_found := found.__len__()) == 0:
        output = "Unidentified specification file"
    elif n_found == 1:
        output = found[0]
    else:
        alternatives = ", ".join(found)
        output = f"Alternatives: {alternatives}"

    return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import sys as sstm
from argparse import ArgumentParser as argument_parser_t

from str_to_obj import ObjectFromStr

from conf_ini_g.extension.path import (
    ValidateInputPath,
    ValidateOutputPath,
    any_path_h,
    path_t,
)
from conf_ini_g.phase.raw.config import AsStr, NewRawConfig, any_raw_config_h
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.section.main import section_t
from conf_ini_g.phase.specification.section.unit import IsUnitSection

_SECTIONS = "SECTIONS"


def DraftSpecificationFromINIDocument(path: any_path_h, /) -> str | None:
    """"""
    ini_config = NewRawConfig(ini_path=path)
    if ini_config is None:
        return None

    sections = []
    for section_name, parameters in ini_config.items():
        # possibly_fuzzy=True: in case the raw config is not valid in that respect
        if IsUnitSection(section_name, possibly_fuzzy=True):
            continue

        parameters_as_lst = []
        for parameter_name, value_as_str in parameters.items():
            value, _ = ObjectFromStr(value_as_str)
            parameter = (
                f"{parameter_t.__name__}(\n"
                f'                name="{parameter_name}",\n'
                f"                type={type(value).__name__},\n"
                f"            )"
            )
            parameters_as_lst.append(parameter)

        parameters_as_str = ",\n            ".join(parameters_as_lst)
        section = (
            f"    {section_t.__name__}(\n"
            f'        name="{section_name}",\n'
            f"        parameters=[\n"
            f"            {parameters_as_str}\n"
            f"        ]\n"
            f"    )"
        )
        sections.append(section)

    imports = (
        f"# To use this specification file:\n"
        f"#     1. import the object {_SECTIONS}\n"
        f"#     2. instantiate a conf_ini_g.specification.config.config_t from it\n"
        f"from conf_ini_g.specification.parameter.main "
        f"import {parameter_t.__name__}\n"
        f"from conf_ini_g.specification.section.main import {section_t.__name__}\n"
    )

    return imports + f"\n{_SECTIONS} = (\n" + ",\n".join(sections) + ",\n)\n"


def SaveRawConfigToINIDocument(
    config: any_raw_config_h,
    path: any_path_h,
    /,
    *,
    should_overwrite: bool = False,
    should_raise_on_error: bool = False,
) -> list[str]:
    """"""
    path = path_t(path)
    issues = ValidateOutputPath(
        path,
        should_overwrite=should_overwrite,
        should_raise_on_error=should_raise_on_error,
    )
    if issues.__len__() > 0:
        return issues

    encoding = sstm.getfilesystemencoding()

    with path.open("w", encoding=encoding) as ini_writer:
        print(AsStr(config), file=ini_writer)
        print(f"", file=ini_writer)

    return []


if __name__ == "__main__":
    """
    Run with: python -m conf_ini_g.interface.storage.config
    from package base folder.
    """
    main_encoding = sstm.getfilesystemencoding()

    parser = argument_parser_t(
        prog="python -m conf_ini_g.interface.storage.config",
        description="Display or save a draft config specification "
        "based on an INI file.",
        allow_abbrev=False,
    )
    # type=argparse.FileType() => automatic file opening
    parser.add_argument(
        dest="ini_path",
        help="Input: Path to INI configuration file",
        metavar="INI_config_file",
    )
    # type=argparse.FileType('w', encoding=main_encoding) => automatic creation of file
    parser.add_argument(
        dest="draft_specification",
        help="Output: Path to draft config specification file. "
        "If not passed, specification is displayed in console.",
        default=None,
        nargs="?",
        metavar="draft_spec_file",
    )
    parser.add_argument(
        "--overwrite",
        dest="should_overwrite",
        action="store_true",
        help="Allows overwriting of draft config specification file.",
    )

    arguments = parser.parse_args()
    ini_path = path_t(arguments.ini_path)
    draft_path = arguments.draft_specification
    main_should_overwrite = arguments.should_overwrite

    issues_ = ValidateInputPath(ini_path)
    if issues_.__len__() > 0:
        print("\n".join(issues_) + "\n")
        parser.print_help()
        sstm.exit(-1)

    if draft_path is not None:
        draft_path = path_t(draft_path)
        issues_ = ValidateOutputPath(draft_path, should_overwrite=main_should_overwrite)
        if issues_.__len__() > 0:
            print("\n".join(issues_) + "\n")
            parser.print_help()
            sstm.exit(-1)

    draft = DraftSpecificationFromINIDocument(ini_path)
    if draft is None:
        print(f"{ini_path}: Not a valid INI document")
        sstm.exit(-1)

    if draft_path is None:
        print(draft)
    else:
        with draft_path.open("w", encoding=main_encoding) as draft_writer:
            draft_writer.write(draft)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import configparser as cfpr
import sys as sstm
from typing import Any, Literal, get_args

from conf_ini_g.extension.path import any_path_h
from conf_ini_g.extension.string import Flattened
from conf_ini_g.interface.constant import (
    INI_COMMENT_MARKER,
    INI_VALUE_ASSIGNMENT,
    PARAMETER_COLOR,
    SECTION_COLOR,
)

ini_config_h = dict[str, dict[str, str]]  # Without value interpretation
typed_config_h = dict[
    str, dict[str, Any]
]  # With interpreted values, and possibly units
any_raw_config_h = ini_config_h | typed_config_h


color_modes_h = Literal["rich", "html", None]


def NewRawConfig(
    *,
    ini_path: any_path_h | None = None,
    arguments: typed_config_h = None,
) -> ini_config_h:
    """"""
    if ini_path is None:
        ini_config = {}
    else:
        ini_config = cfpr.ConfigParser(
            delimiters=INI_VALUE_ASSIGNMENT,
            comment_prefixes=INI_COMMENT_MARKER,
            empty_lines_in_values=False,
            interpolation=None,
        )
        ini_config.optionxform = lambda option: option
        # Returns DEFAULT <Section: DEFAULT> if path does not exist or is a folder.
        ini_config.read(ini_path, encoding=sstm.getfilesystemencoding())

    output = {
        section: {parameter: value for parameter, value in parameters.items()}
        for section, parameters in ini_config.items()
        if section != cfpr.DEFAULTSECT
    }
    if arguments is not None:
        for sct_name, parameters in arguments.items():
            if sct_name not in output:
                output[sct_name] = {}
            for prm_name, value in parameters.items():
                output[sct_name][prm_name] = value

    return output


def AsStr(
    config: any_raw_config_h,
    /,
    *,
    color: color_modes_h = None,
    section_color: str = None,
    parameter_color: str = None,
) -> str:
    """"""
    output = []

    if section_color is None:
        section_color = SECTION_COLOR
    if parameter_color is None:
        parameter_color = PARAMETER_COLOR

    if color is None:
        section_color = ""
        parameter_color = ""
        color_reset = ""
        newline = "\n"
        indentation = " "
        bracket_escape = ""
    elif color == "rich":
        section_color = f"[{section_color}]"
        parameter_color = f"[{parameter_color}]"
        color_reset = "[/]"
        newline = "\n"
        indentation = " "
        bracket_escape = "\\"
    elif color == "html":
        section_color = f'<span style="color:{section_color}">'
        parameter_color = f'<span style="color:{parameter_color}">'
        color_reset = "</span>"
        newline = "<br/>"
        indentation = "&nbsp;"
        bracket_escape = ""
    else:
        raise ValueError(
            f"{color}: Invalid color mode; Expected=one of {get_args(color_modes_h)}."
        )

    longest = 0
    for section, parameters in config.items():
        if parameters.__len__() == 0:
            continue

        inner_output = []
        lengths = []
        for name, value in parameters.items():
            length = name.__len__()
            lengths.append(length)
            longest = max(longest, length)

            flattened = Flattened(str(value))
            inner_output.append(f"{parameter_color}{name}{color_reset}@= {flattened}")

        output.append(
            (
                f"{section_color}{bracket_escape}[{section}]{color_reset}",
                inner_output,
                lengths,
            )
        )

    output = (
        f"{_sct}{newline}"
        + newline.join(
            _lne.replace("@", (longest - _lgt + 1) * indentation, 1)
            for _lne, _lgt in zip(_prm, _lgs)
        )
        for _sct, _prm, _lgs in output
    )

    return f"{newline}{newline}".join(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import sys as sstm
import textwrap as text
from typing import Any, Sequence

from rich import print as rprint
from rich.text import Text as text_t
from str_to_obj import ObjectFromStr
from str_to_obj.interface.console import TypeAsRichStr
from str_to_obj.type.hint import any_hint_h

from conf_ini_g.catalog.specification.parameter.choices import choices_t
from conf_ini_g.extension.python import SpecificationPath
from conf_ini_g.interface.constant import INI_UNIT_SECTION, INI_UNIT_SEPARATOR
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.parameter.type import type_t
from conf_ini_g.phase.specification.parameter.unit import unit_t
from conf_ini_g.phase.specification.parameter.value import MISSING_REQUIRED_VALUE
from conf_ini_g.phase.specification.section.main import controller_t, section_t
from conf_ini_g.phase.specification.section.unit import IsUnitSection


@dtcl.dataclass(init=False, repr=False, eq=False)
class config_t(list[section_t]):
    spec_path: str = None

    def __init__(self, sections: Sequence[section_t], /) -> None:
        """
        Raising exceptions is adapted here since execution cannot proceed without a
        valid specification.
        """
        list.__init__(self, sections)

        # After __init__ so that self iterator is usable
        for section in self:
            if (controller := section.controller) is None:
                continue

            choices = choices_t.NewAnnotatedType(section.controlling_values)
            self.GetController(controller).type = type_t.NewFromTypeHint(choices)

        self.spec_path = SpecificationPath(sections)

        issues = self.Issues()
        if issues.__len__() > 0:
            rprint(
                self.spec_path,
                "[red]Invalid specification[/]",
                "\n".join(issues),
                sep="\n",
            )
            sstm.exit(1)

    def AddUnitSection(self) -> None:
        """"""
        section = section_t(
            name=INI_UNIT_SECTION,
            definition="Unit definitions",
            description=f"Units can be used in any other section "
            f"to specify a parameter value as follows: "
            f"numerical_value{INI_UNIT_SEPARATOR}unit, e.g., 1.5'mm.",
            basic=True,
            optional=True,
            category=INI_UNIT_SECTION,
            is_growable=True,
        )
        self.append(section)

    @property
    def section_names(self) -> Sequence[str]:
        """"""
        return tuple(_sct.name for _sct in self)

    def AddLateParameter(
        self,
        section: str | section_t,
        name: str,
        /,
        *,
        definition: str = 'Programmatic "late" parameter',
        description: str = "This parameter is not part of the specification. "
        'It was added programmatically in a "plugin" way.',
        basic: bool = True,
        type_: any_hint_h | type_t = None,
        default: Any = MISSING_REQUIRED_VALUE,
        controlling_value: str = None,
    ) -> None:
        """
        See definition and description above.
        """
        if isinstance(section, str):
            section = self[section]

        parameter = parameter_t(
            name=name,
            definition=definition,
            description=description,
            basic=basic,
            type=type_,
            default=default,
        )
        self._AddParameterInSection(
            section, parameter, controlling_value, config_t.AddLateParameter.__name__
        )

    def AddINIParameter(
        self,
        section: str | section_t,
        name: str,
        value: str,
        /,
        *,
        controlling_value: str = None,
    ) -> parameter_t:
        """
        See definition and description below.
        The existence of such a method is justified by the fact that the parameter
        created can be a "normal" parameter or a unit depending on the section.
        """
        if isinstance(section, str):
            section = self[section]

        if IsUnitSection(section.name):
            parameter_or_unit_t = unit_t
            basic = True
            converted = None
            type_ = None  # Correctly set by the constructor.
        else:
            parameter_or_unit_t = parameter_t
            basic = section.basic
            converted, _ = ObjectFromStr(value)
            type_ = type(converted)
        definition = "Programmatic INI/Cmd parameter"
        description = (
            "This parameter is not part of the specification. "
            "It was added programmatically because it was found in the INI document, "
            "or passed as a command-line argument."
        )
        parameter = parameter_or_unit_t(
            name=name,
            definition=definition,
            description=description,
            basic=basic,
            type=type_,
            default=converted,  # Just a trick to prevent error if basic is False.
        )
        self._AddParameterInSection(
            section, parameter, controlling_value, config_t.AddINIParameter.__name__
        )

        return parameter

    def _AddParameterInSection(
        self,
        section: section_t,
        parameter: parameter_t,
        controlling_value: str,
        caller: str,
        /,
    ) -> None:
        """"""
        if (controller := section.controller) is None:
            section.append(parameter)
            return

        if controlling_value is None:
            raise ValueError(
                f"{caller}: A Controlling value must be passed for parameter "
                f"{section.name}.{parameter.name}."
            )

        if controller.primary_value is None:
            # controller_t is immutable, so it must be re-created.
            section.controller = controller_t(
                section=controller.section,
                parameter=controller.parameter,
                primary_value=controlling_value,
            )
            controller = section.controller
            section.append(parameter)
            should_update_controller_choices = True
        elif controlling_value == controller.primary_value:
            section.append(parameter)
            should_update_controller_choices = False
        elif section.alternatives is None:
            section.alternatives = {controlling_value: [parameter]}
            should_update_controller_choices = True
        elif controlling_value in section.alternatives:
            section.alternatives[controlling_value].append(parameter)
            should_update_controller_choices = False
        else:
            section.alternatives[controlling_value] = [parameter]
            should_update_controller_choices = True

        if should_update_controller_choices:
            controlling_values = section.controlling_values + (controlling_value,)
            choices = choices_t.NewAnnotatedType(controlling_values)
            self.GetController(controller).type = type_t.NewFromTypeHint(choices)

    def GetController(self, controller: controller_t, /) -> parameter_t:
        """"""
        return self[controller.section][controller.parameter]

    def Issues(self) -> list[str]:
        """"""
        output = []

        if self.__len__() == 0:
            output.append("spec_path: Empty specification")
            return output

        names = self.section_names
        if names.__len__() > set(names).__len__():
            output.append("Specification with repeated section names")

        for section in self:
            if not isinstance(section, section_t):
                output.append(
                    f"{type(section).__name__}: Invalid section type; Expected={section_t.__name__}."
                )
                continue

            output.extend(section.Issues())
            if section.controller is not None:
                if section.controller.section not in self:
                    output.append(
                        f"{section.controller.section}: "
                        f'Unspecified section declared as controller of section "{section.name}"'
                    )
                else:
                    controller_section = self[section.controller.section]
                    if controller_section.controller is not None:
                        output.append(
                            f"{section.controller.section}: "
                            f'Section controlling "{section.name}" is itself controlled'
                        )
                    if section.controller.parameter not in controller_section:
                        output.append(
                            f"{section.controller.section}.{section.controller.parameter}: "
                            f'Unspecified parameter declared as controller of section "{section.name}"'
                        )
                    else:
                        controller_parameter = controller_section[
                            section.controller.parameter
                        ]
                        if controller_parameter.optional:
                            output.append(
                                f"{section.controller.section}.{section.controller.parameter}: "
                                f'Optional parameter declared as controller of section "{section.name}"'
                            )

        return output

    def _Item(self, key: str | int, /) -> section_t | None:
        """"""
        if isinstance(key, int):
            return list.__getitem__(self, key)

        for section in self:
            if section.name == key:
                return section

        return None

    def __contains__(self, key: str, /) -> bool:
        """"""
        return self._Item(key) is not None

    def __getitem__(self, key: str | int, /) -> section_t:
        """"""
        item = self._Item(key)
        if item is None:
            raise KeyError(f"{key}: Not a section of config.")

        return item

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [
            TypeAsRichStr(self),
            f"    [blue]spec_path[/]={self.spec_path}"
            f"[yellow]:{type(self.spec_path).__name__}[/]",
        ]

        for section in self:
            output.append(text.indent(section.__rich__(), "    "))

        return "\n".join(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Callable, Sequence

from babelwidget.main import backend_t, button_wgt_h, menu_wgt_h, vbox_lyt_h

from conf_ini_g.catalog.interface.window.parameter.path import path_wgt_t
from conf_ini_g.catalog.specification.parameter.path import path_t as path_annotation_t
from conf_ini_g.phase.instance.config.main import config_t as config_instance_t


def TitleLayout(
    title: str | None,
    instance: config_instance_t,
    history: Sequence[str] | None,
    backend: backend_t,
    UpdateWithNewINI: Callable,
    Close: Callable,
    /,
) -> tuple[vbox_lyt_h, path_wgt_t, button_wgt_h, menu_wgt_h, button_wgt_h]:
    """"""
    if title is None:
        title = "Conf-INI-g"

    layout = backend.hbox_lyt_t()
    inner_layout = backend.vbox_lyt_t()

    title = (
        f'<h1 style="color: blue">{title}</h1>'
        f"<b><font face=monospace>SPEC:</font></b> {instance.specification.spec_path}"
    )
    title_wgt = backend.label_wgt_t(title)
    title_wgt.setAlignment(backend.ALIGNED_LEFT)

    label_wgt = backend.label_wgt_t("<b><font face=monospace>INI: </font></b>")
    ini_path_wgt = path_wgt_t.NewWithDetails(
        instance.ini_path,
        path_annotation_t(
            target_type=path_annotation_t.TARGET_TYPE.document, is_input=True
        ),
        backend,
        None,
        editable=False,
        PostAssignmentFunction=UpdateWithNewINI,
    )
    history_button = backend.button_wgt_t("🕑")
    history_menu = backend.menu_wgt_t.NewWithFunction(UpdateWithNewINI)
    if history is None:
        history_button.setEnabled(False)
    else:
        for choice in history:
            history_menu.addAction(choice)
    history_button.setMenu(history_menu)
    #
    label_wgt.setSizePolicy(backend.SIZE_FIXED, backend.SIZE_FIXED)
    history_button.setSizePolicy(backend.SIZE_FIXED, backend.SIZE_MINIMUM)
    #
    path_layout = backend.hbox_lyt_t()
    path_layout.addWidget(label_wgt)
    path_layout.addWidget(ini_path_wgt.library_wgt)
    path_layout.addWidget(history_button)

    inner_layout.addWidget(title_wgt)
    inner_layout.addLayout(path_layout)

    close_button = backend.button_wgt_t("CLOSE")
    close_button.SetFunction(Close)
    close_button.setSizePolicy(backend.SIZE_FIXED, backend.SIZE_MINIMUM)

    layout.addLayout(inner_layout)
    layout.addWidget(close_button)

    return layout, ini_path_wgt, history_button, history_menu, close_button



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl

from babelwidget.main import backend_t, choices_dots_wgt_h, library_wgt_h

from conf_ini_g.catalog.specification.parameter.boolean import boolean_t
from conf_ini_g.interface.window.component.parameter.protocol import parameter_a
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class boolean_wgt_t(parameter_a):
    library_wgt: library_wgt_h
    true_btn: choices_dots_wgt_h = dtcl.field(init=False, default=None)

    @classmethod
    def NewWithDetails(
        cls,
        value: bool | None,
        value_type: type_t | boolean_t | None,
        backend: backend_t,
        _: parameter_t | None,
        /,
    ) -> boolean_wgt_t:
        """
        If value_type does not contain the necessary details, an exclamation point is added to the default values.
        """
        output = cls(library_wgt=backend.library_wgt_t())

        value = _ValidValue(value)

        if value_type is None:
            labels = None
        else:
            if isinstance(value_type, type_t):
                annotation = value_type.FirstAnnotationWithAttribute("mode")
            else:
                annotation = value_type
            if annotation is None:
                labels = None
            else:
                labels = getattr(annotation.mode, "value", None)
        if labels is None:
            labels = ("True!", "False!")

        true_btn = backend.dot_button_wgt_t(labels[0], parent=output.library_wgt)
        false_btn = backend.dot_button_wgt_t(labels[1], parent=output.library_wgt)
        true_btn.setChecked(value)
        false_btn.setChecked(not value)

        output.true_btn = true_btn

        layout = backend.hbox_lyt_t()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(true_btn)
        layout.addWidget(false_btn)
        output.library_wgt.setLayout(layout)

        return output

    def Assign(self, value: bool | None, _: type_t | boolean_t | None, /) -> None:
        """"""
        self.true_btn.setChecked(_ValidValue(value))

    def Text(self) -> str:
        """"""
        return str(self.true_btn.isChecked())


def _ValidValue(value: bool | None, /) -> bool:
    """"""
    if value is None:
        return False
    else:
        return value



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from abc import ABC as abstract_class_t
from abc import abstractmethod
from typing import Any

from rich.text import Text as text_t

from conf_ini_g.phase.instance.parameter.value import INVALID_VALUE, invalid_value_t
from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class base_t(abstract_class_t):
    # str: From INI document or interfaces.
    # Any: For default values.
    # None: For empty text in interfaces.
    value: str | Any | None = None
    comment: str | None = None

    @abstractmethod
    def SetINIorInterfaceOrDefaultValue(
        self,
        value_w_unit_w_comment: str | Any,
        comment_marker: str,
        /,
        *,
        unit: str = None,
    ) -> None:
        """"""
        ...

    @abstractmethod
    def TypedValue(
        self,
        expected_type: type_t,
        /,
        *,
        units: dict[str, int | float | invalid_value_t] | None = None,
    ) -> tuple[Any, list[str]]:
        """"""
        ...

    @abstractmethod
    def Text(self) -> str:
        """"""
        ...

    def Issues(self) -> list[str]:
        """"""
        if self.value is INVALID_VALUE:
            return ["Invalid value"]

        return []

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        return self.Text()


def Pieces(
    combined: str, separator: str, /, *, from_left: bool = True
) -> tuple[str, str | None]:
    """"""
    if from_left:
        where_separator = combined.find(separator)
    else:
        where_separator = combined.rfind(separator)

    if where_separator != -1:
        left = combined[:where_separator].strip()
        right = combined[(where_separator + separator.__len__()) :].strip()
    else:
        left = combined
        right = None

    return left, right



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re as rgex
from pathlib import Path as path_t

from setuptools import setup

HERE = path_t(__file__).parent.resolve()
DOCUMENTATION_HOME = HERE / "documentation" / "wiki" / "description.asciidoc"


def DescriptionFromDocumentation(documentation: path_t, /) -> dict[str, str]:
    """"""
    output = {}

    pattern = rgex.compile(r":([A-Z_]+): +(.+)\n?", flags=rgex.ASCII)

    with open(documentation) as accessor:
        for line in accessor.readlines():
            if (match := pattern.fullmatch(line)) is not None:
                name = match.group(1)
                value = match.group(2)
                output[name] = value

    return output


DESCRIPTION = DescriptionFromDocumentation(DOCUMENTATION_HOME)


LICENSE_SHORT = "CeCILL-2.1"
LICENCE_LONG = "CEA CNRS Inria Logiciel Libre License, version 2.1"
PY_VERSION = "3.10"

DOCUMENTATION_SITE = "-/wikis/home"

PYPI_NAME = "conf-ini-g"
PYPI_TOPIC = "Software Development"
PYPI_AUDIENCE = "Developers"
PYPI_STATUS = "4 - Beta"

IMPORT_NAME = "conf_ini_g"
PACKAGES = [
    IMPORT_NAME,
    f"{IMPORT_NAME}.catalog",
    f"{IMPORT_NAME}.catalog.interface",
    f"{IMPORT_NAME}.catalog.interface.window",
    f"{IMPORT_NAME}.catalog.interface.window.parameter",
    f"{IMPORT_NAME}.catalog.specification",
    f"{IMPORT_NAME}.catalog.specification.parameter",
    f"{IMPORT_NAME}.extension",
    f"{IMPORT_NAME}.interface",
    f"{IMPORT_NAME}.interface.console",
    f"{IMPORT_NAME}.interface.window",
    f"{IMPORT_NAME}.interface.window.component",
    f"{IMPORT_NAME}.interface.window.component.config",
    f"{IMPORT_NAME}.interface.window.component.parameter",
    f"{IMPORT_NAME}.interface.window.component.section",
    f"{IMPORT_NAME}.interface.storage",
    f"{IMPORT_NAME}.light",
    f"{IMPORT_NAME}.phase",
    f"{IMPORT_NAME}.phase.instance",
    f"{IMPORT_NAME}.phase.instance.config",
    f"{IMPORT_NAME}.phase.instance.parameter",
    f"{IMPORT_NAME}.phase.raw",
    f"{IMPORT_NAME}.phase.specification",
    f"{IMPORT_NAME}.phase.specification.config",
    f"{IMPORT_NAME}.phase.specification.parameter",
    f"{IMPORT_NAME}.phase.specification.section",
]
EXCLUDED_FOLDERS = (
    f"{IMPORT_NAME}.documentation",
    f"{IMPORT_NAME}.test",
)
ENTRY_POINTS = {}


long_description = (HERE / "README.rst").read_text(encoding="utf-8")
repository_url = (
    f"https://"
    f"{DESCRIPTION['REPOSITORY_SITE']}/"
    f"{DESCRIPTION['REPOSITORY_USER']}/"
    f"{DESCRIPTION['REPOSITORY_NAME']}/"
)
documentation_url = f"{repository_url}/{DOCUMENTATION_SITE}"


def CheckCoherenceBetweenDeclarationAndReality() -> None:
    """"""
    folders = [IMPORT_NAME]
    for node in (HERE / IMPORT_NAME).rglob("*"):
        if node.is_dir() and not str(node).startswith("."):
            node = node.relative_to(HERE)
            node = ".".join(node.parts)
            if not (
                (node in EXCLUDED_FOLDERS)
                or any(node.startswith(_fld + ".") for _fld in EXCLUDED_FOLDERS)
            ):
                folders.append(node)
    folders = sorted(folders)

    packages = sorted(PACKAGES)
    if packages != folders:
        raise ValueError(
            f"Mismatch between declared and found packages:\n"
            f"    - Declared=\n      {packages}\n"
            f"    - Found=\n      {folders}\n"
            f"    - Undeclared=\n      {set(folders).difference(packages)}\n"
            f"    - Nonexistent=\n      {set(packages).difference(folders)}"
        )


def Version():
    """"""
    contents = {}
    with open(HERE / IMPORT_NAME / "version.py") as accessor:
        exec(accessor.read(), contents)

    output = contents["__version__"]
    if isinstance(output, str) and rgex.fullmatch(r"20[0-9]{2}\.[1-9][0-9]*", output):
        return output

    raise ValueError(f"{output}: Invalid version.")


if __name__ == "__main__":
    #
    CheckCoherenceBetweenDeclarationAndReality()
    # fmt: off
    setup(
        author=DESCRIPTION["AUTHOR"],
        author_email=DESCRIPTION["EMAIL"],
        #
        name=PYPI_NAME,
        description=DESCRIPTION["SHORT_DESCRIPTION"],
        long_description=long_description,
        long_description_content_type="text/x-rst",
        license=LICENSE_SHORT,
        version=Version(),
        #
        classifiers=[
            f"Topic :: {PYPI_TOPIC}",
            f"Intended Audience :: {PYPI_AUDIENCE}",
            f"License :: OSI Approved :: {LICENCE_LONG} ({LICENSE_SHORT})",
            f"Programming Language :: Python :: {PY_VERSION}",
            f"Development Status :: {PYPI_STATUS}",
        ],
        keywords=DESCRIPTION["KEYWORDS"],
        #
        url=repository_url,
        project_urls={
            "Documentation": documentation_url,
            "Source": repository_url,
        },
        #
        packages=PACKAGES,
        entry_points=ENTRY_POINTS,
        python_requires=f">={PY_VERSION}",
        install_requires=[
            "babelwidget",
            "openpyxl",
            "rich",
            "str-to-obj",
        ],
    )



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from typing import Any

from conf_ini_g.interface.constant import INI_UNIT_SEPARATOR
from conf_ini_g.phase.instance.parameter.base import Pieces, base_t
from conf_ini_g.phase.instance.parameter.unit import ConvertedValue
from conf_ini_g.phase.instance.parameter.value import INVALID_VALUE, invalid_value_t
from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class instance_t(base_t):
    unit: str = None

    def SetINIorInterfaceOrDefaultValue(
        self,
        value_w_unit_w_comment: str | Any,
        comment_marker: str,
        /,
        *,
        unit: str = None,
    ) -> None:
        """"""
        value_as_str, inline_unit, comment = _ValueUnitAndComment(
            value_w_unit_w_comment, comment_marker
        )
        if unit is None:
            unit = inline_unit

        self.value = value_as_str
        self.unit = unit
        self.comment = comment

    def TypedValue(
        self,
        expected_type: type_t,
        /,
        *,
        units: dict[str, int | float | invalid_value_t] = None,
    ) -> tuple[Any, list[str]]:
        """
        With unit consumed or not.
        """
        final_value, issues = expected_type.InterpretedValueOf(self.value)
        if issues.__len__() > 0:
            return INVALID_VALUE, [f"{self.value}: Invalid value: {', '.join(issues)}"]

        if (units is None) or ((unit := self.unit) is None):
            return final_value, []

        conversion_factor = units.get(unit, None)
        if conversion_factor is None:
            return INVALID_VALUE, [f"{unit}: Missing unit definition."]
        if conversion_factor is INVALID_VALUE:
            return INVALID_VALUE, [f"{unit}: Invalid unit value."]

        converted, unconverted = ConvertedValue(final_value, conversion_factor)
        if unconverted.__len__() > 0:
            unconverted = ", ".join(unconverted)
            return INVALID_VALUE, [
                f"{unconverted}: Value(s) do(es) not support unit conversion."
            ]

        return converted, []

    def Text(self) -> str:
        """"""
        if self.unit is None:
            return str(self.value)  # str: Only useful when default value.

        return f"{self.value}{INI_UNIT_SEPARATOR}{self.unit}"


def _ValueUnitAndComment(
    value_w_unit_w_comment: str,
    comment_marker: str,
    /,
) -> tuple[str, str | None, str | None]:
    """"""
    value_w_unit, comment = Pieces(value_w_unit_w_comment, comment_marker)
    if (comment is not None) and (comment.__len__() == 0):
        comment = None

    value_as_str, unit = Pieces(value_w_unit, INI_UNIT_SEPARATOR, from_left=False)
    # if unit.__len__() == 0, do not make it None so that an invalid unit error is noticed later on

    return value_as_str, unit, comment



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from math import pi as PI

INI_UNIT_SECTION = "UNITS"  # In uppercase

INI_VALUE_ASSIGNMENT = "="
INI_UNIT_SEPARATOR = ":IN:"
INI_COMMENT_MARKER = "#"


_STD_UNIT_CONVERSIONS = (
    # unit, unit name, parent unit (None if none), conversion factor (1.0 if None)
    ("si", "site", None, 1.0),  # Synonym for sample, pixel, voxel...
    #
    ("km", "kilometer", "m", 1000.0),
    ("m", "meter", "si", 1.0),
    ("cm", "centimeter", "m", 1.0e-2),
    ("mm", "millimeter", "m", 1.0e-3),
    ("um", "micrometer", "m", 1.0e-6),
    ("nm", "nanometer", "m", 1.0e-9),
    ("mi", "mile", "m", 1.0 / 0.00062137),
    ("yd", "yard", "m", 0.9144),
    ("ft", "foot", "m", 1.0 / 3.2808),
    ("in", "inch", "m", 1.0 / 39.37),
    #
    ("r", "radian", "si", 1.0),
    ("d", "degree", "r", PI / 180.0),
    #
    ("wk", "week", "s", 604800.0),
    ("dy", "day", "s", 86400.0),
    ("h", "hour", "s", 3600.0),
    ("mn", "minute", "s", 60.0),
    ("s", "second", "si", 1.0),
    ("ms", "millisecond", "s", 1.0e-3),
    ("us", "microsecond", "s", 1.0e-6),
    ("ns", "nanosecond", "s", 1.0e-9),
)
STD_UNIT_CONVERSIONS = {_elm[0]: _elm[-1] for _elm in _STD_UNIT_CONVERSIONS}


STD_ABBREVIATIONS = {
    "fct": "function",
    "mod": "module",
    "n": "number of",
    "o": "of",
    "prm": "parameter",
    "rng": "range",
}


SECTION_COLOR = "green"
PARAMETER_COLOR = "blue"


PATH_SELECTOR_WIDTH = 100
